/**********************************************************************
 * File Name: MobileObject.java
 *
 * Copyright (C) 2000, US Air Force Research Laboratory,
 * Rome Research Site, Rome NY 13424, USA.
 *
 * Use of this software implies acceptance of the terms described in
 * accompanying license.txt file.
 *
 *
 * Authors: John Crowter, Jennifer Dzek, James Lawton, 
 *          Bob Paragi, and Joe Beasock
 *
 * Creation Date: May 2000
 *
 * Description:
 *
 * Modification History:
 *
 *  - June 00, Bryan Horling, UMass (BCH): Changes to make drawing more 
 *  efficient by only redrawing when needed.
 *
 **********************************************************************/

import java.lang.*;
import java.math.*;
import java.awt.*;
import java.util.*;
import java.io.*;

public class MobileDirectedObject extends MobileObject {

    // way_points holds the list of way points with the target's starting
    // location being way_points[0] = x_coord and way_points[1] = y_coord
    // the array list is a list of numbers of type double and of the form
    // [x1,y1,x2,y2,....,xn,yn]
    protected ArrayList way_points;
    // next_x and next_y hold the (x,y) location of the next way point to be
    // encountered. The initial values are 2 and 3 respectively since the
    // first way point, in 0 and 1, are the x_coord and y_coord, respectively
    protected int next_x, next_y;


    public MobileDirectedObject() {
	name = new String("Unknown");
    }

    public MobileDirectedObject (String args, double width, double height)
	throws Exception {
	StringTokenizer s = new StringTokenizer(args);
	int n = s.countTokens();
	boolean bad_data = false;
	Double x, y;
	way_points = new ArrayList();
	next_x = 2;
	next_y = 3;

	// n must be equal to or greater than 8 to account for at least
	// one more point in additon to the starting point
	// AND are an even number of tokens
	if ((n >= 7) && (1 == (n % 2))) {// at least TWO points present
	    s.nextToken();               // Get rid of the "MOBOBJ" tag
	    name = s.nextToken();
	    x_coord = (new Double(s.nextToken())).doubleValue();
	    way_points.add(new Double(x_coord));
	    if ((x_coord < 0.0) || (x_coord > width))
		bad_data = true;
	    y_coord = (new Double(s.nextToken())).doubleValue();
	    way_points.add(new Double(y_coord));
	    if ((y_coord < 0.0) || (y_coord > height))
		bad_data = true;
	    speed = (new Double(s.nextToken())).doubleValue();
	    if (speed < 0.0)
		bad_data = true;
	    // mdirection = (new Double(s.nextToken())).doubleValue();
	    // if ((mdirection < 0.0) || (mdirection > 360.0))
	    //  bad_data = true;

	    // for each remaining pair of (x,y) coordinates, add
	    // them to the way_points list and make sure they're
	    // within range
	    while (s.hasMoreTokens()) { // add the rest of the way points
		x = (new Double(s.nextToken()));
		way_points.add(x);
		if ((x.doubleValue() < 0.0) || (x.doubleValue() > width))
		    bad_data = true;
		y = (new Double(s.nextToken()));
		way_points.add(y);
		if ((y.doubleValue() < 0.0) || (y.doubleValue() > height))
		    bad_data = true;
	    }

	    //System.out.println("Waypoints: " + way_points.toString());

	    x = (Double) way_points.get(next_x);
	    y = (Double) way_points.get(next_y);
	    // compute the resulting new 'mdirection'
	    mdirection = new_direction(x.doubleValue() - x_coord,y.doubleValue() - y_coord);
      
	    scanned = false;
	}
	else throw (new Exception("Too few args"));
	if (bad_data) throw (new Exception("argument(s) out of range"));
    }

    public double distance(double x, double y, double a, double b) {
	double xa = (x - a) * (x - a);
	double yb = (y - b) * (y - b);
	return Math.sqrt((xa + yb));
    }

    public double get_wp_coord(int index){
	Double wp = (Double) way_points.get(index);
	return wp.doubleValue();
    }


    // Update our location base on how long its been since the
    // last location update.
    public void next_loc (int simtime) {
	double distance,total_distance, time, radian, sx, sy,nx, ny, tx, ty;

	int ox = env.scale_x_coord ( x_coord, (double) lastw),
	    oy = env.scale_y_coord ( y_coord, (double) lasth); // BCH

	// Compute time in seconds
	time = (simtime - last_update) * ((double) env.clock.ticksize / 1000.0);
	last_update = simtime;

	if (time > 0) {
	    radian = (mdirection) * (Math.PI/180);
	    nx = x_coord + (time * speed * Math.cos(radian));
	    ny = y_coord + (time * speed * Math.sin(radian));
	    // the total distance the target has traveled since the last update is ...
	    total_distance = distance(x_coord,y_coord,nx,ny);
	    // the distance from the start position of the target to the next way point is ...
	    distance = distance(x_coord,y_coord,get_wp_coord(next_x),get_wp_coord(next_y));

	    while (distance < total_distance) {
		// update the target's current position to the next way point
		x_coord = get_wp_coord(next_x);
		y_coord = get_wp_coord(next_y);
		// update where to look in the array for the next way point ...
		next_x = (next_x + 2) % way_points.size();
		next_y = next_x + 1;
		// subtract the distance from the current positon to the next way point
		// from the total distnace to be traveled
		total_distance = total_distance - distance;
		// compute the distance of the next leg and begin the while loop again
		distance = distance(x_coord,y_coord,get_wp_coord(next_x),get_wp_coord(next_y));
	    }

	    // transpose the current leg to the origin and ..
	    nx = get_wp_coord(next_x) - x_coord;
	    ny = get_wp_coord(next_y) - y_coord;
	    // compute the resulting new 'mdirection'
	    mdirection = new_direction(nx,ny);
	    radian = (mdirection) * (Math.PI/180);
	    // compute the final (x,y) location of the target
	    x_coord = x_coord + (total_distance * Math.cos(radian));
	    y_coord = y_coord + (total_distance * Math.sin(radian));

	}
	if (env.Debug > 0) {
	    System.out.println ("Mobile Object " + name + " new location: (" 
				+ x_coord + "," + y_coord + ")");
	    // System.out.println ("  Local Radian value: " + radian);
	}

	// Only redraw graphics when needed (BCH)
	if ((ox != env.scale_x_coord ( x_coord, (double) lastw)) ||
	    (oy != env.scale_y_coord ( y_coord, (double) lasth)))
	    env.requestRedraw();
    }
}
